@@ -26,10 +26,13 @@ char homedir[FILENAME_MAX];
 char chrootdir[FILENAME_MAX];
 char *safeenv[MAX_ENV];
 
+/* will point to syslog or a noop */
+void (*debug)(int priority, const char* format, ...);
+
 cmd_t commands[] =
 { 
 #ifdef ENABLE_SFTP
-	{ PROG_SFTP_SERVER, 0 },
+	{ PROG_SFTP_SERVER, 1 },
 #endif /*ENABLE_SFTP*/
 #ifdef ENABLE_SCP2
 	{ PROG_LS, 1 }, 
@@ -78,6 +81,61 @@ cmd_t commands[] =
 	NULL
 };
 
+/*
+ * The array of longopts to be used for validation
+ * longopts := (name, has_args, *flag, val)
+ */
+struct option empty_longopts[] = {
+	{ NULL,			0,	NULL,	0 },
+	};
+
+#ifdef RSYNC_COMPAT
+struct option rsync_longopts[] = {
+	/* I use 'e' for val here because that's what's listed in cmd_arg_t->badarg  */
+	{"rsh", 			1,	0,		(int)'e'},
+	/* the following are disabled because they use daemon mode */
+	{"daemon",			0,	0,		(int)'e'},
+	{"rsync-path",		1,	0,		(int)'e'},
+	{"address",			1,	0,		(int)'e'},
+	{"port",			1,	0,		(int)'e'},
+	{"sockopts",		1,	0,		(int)'e'},
+	{"config",			1,	0,		(int)'e'},
+	{"no-detach",		0,	0,		(int)'e'},
+	{ NULL,				0,	NULL,	0 },
+	};
+#endif
+
+#ifdef SVNSERV_COMPAT
+struct option svnserv_longopts[] = {
+	/* bad */
+	{"daemon",		0,	NULL,	(int)'d' },
+	{"listen-port",	1,	NULL,	(int)'d' },
+	{"listen-host",	1,	NULL,	(int)'d' },
+	{"foreground",	0,	NULL, 	(int)'d' },
+	{"inetd",		0,	NULL, 	(int)'i' },
+	{"threads",		0,	NULL,	(int)'T' },
+	{"listen-once",	0,	NULL,	(int)'X' },
+	/* good */
+	{"read-only",	0,	NULL,	(int)'R' },
+	{"help",		0,	NULL,	(int)'h' },
+	{"root",		0,	NULL,	(int)'r' },
+	{"tunnel",		0,	NULL,	(int)'t' },
+	{"tunnel-user",	1,	NULL,	0 },
+	{ NULL,			0,	NULL,	0 },
+	};
+#endif
+
+#ifdef SVN_COMPAT
+struct option svn_longopts[] = {
+	/* bad */
+	{"editor-cmd",	1,	NULL,	(int)'X' },
+	{"diff-cmd",	1,	NULL,	(int)'X' },
+	{"diff3-cmd",	1,	NULL,	(int)'X' },
+	{"config-dir", 	1,	NULL,	(int)'X' },
+	{ NULL,			0,	NULL,	0 },
+	};
+#endif
+
 /*
  *	several binaries support arbitrary command execution in their arguments
  *	to prevent this we have to check the arguments for these binaries before
@@ -88,20 +146,33 @@ cmd_arg_t dangerous_args[] =
 	/*
 	 *	'oplist' only neccesary where 'use getopt' is 1
 	 *	'strict optlist' only applicable where 'use getopt?' is 1
+	 *	'badarg' is a string to look for if not in rsync mode, if in rsync mode a list of invalid options
 	 *
-	 * program name		use getopt?		strict optlist?	optname			optlist
+	 * program name		use getopt?		strict optlist?	badarg			optlist			longopts\n
 	 */
+#ifdef ENABLE_SFTP
+	{ PROG_SFTP_SERVER,	1,				1,				NULL,			"f:l:",			empty_longopts },
+#endif
 #ifdef ENABLE_SCP2
-	{ PROG_SCP, 		1, 				1,				"S",			"dfl:prtvBCc:i:P:q1246S:o:F:" },
+	{ PROG_SCP, 		1, 				1,				"SoF",			"dfl:prtvBCc:i:P:q1246S:o:F:", empty_longopts },
 #endif
 #ifdef RSYNC_COMPAT
-	{ PROG_RSYNC, 		1, 				0,				"e",			"e:" },
-	{ PROG_RSYNC, 		0, 				0,				"-rsh",			NULL },
-#endif 	
+	{ PROG_RSYNC, 		1, 				0,				"e",			"e:",			rsync_longopts },
+#endif	
 #ifdef UNISON_COMPAT	
-	{ PROG_UNISON, 		0, 				0,				"-rshcmd",		NULL },
-	{ PROG_UNISON, 		0, 				0,				"-sshcmd",		NULL },
-	{ PROG_UNISON, 		0, 				0,				"-servercmd,	NULL" },
+	{ PROG_UNISON, 		0, 				0,				"-rshcmd",		NULL, 			empty_longopts },
+	{ PROG_UNISON, 		0, 				0,				"-sshcmd",		NULL, 			empty_longopts },
+	{ PROG_UNISON, 		0, 				0,				"-servercmd",	NULL, 			empty_longopts },
+	{ PROG_UNISON, 		0, 				0,				"-addversionno",NULL, 			empty_longopts },
+#endif
+#ifdef SVNSERV_COMPAT
+	{ PROG_SVNSERV,		1, 				1,				"diTX",			"dihr:RtTX",	svnserv_longopts },
+#endif
+#ifdef SVN_COMPAT
+	{ PROG_SVN,			1, 				0,				"Xx",			"NvxuRr:qm:F:",		svn_longopts },
+#endif
+#ifdef QUOTA_COMPAT
+	{ PROG_QUOTA,		1,				1,				NULL,			"-F:guvsilqQ",	empty_longopts },
 #endif
 	NULL
 };
@@ -136,6 +207,7 @@ char * allowed_env_vars[] =
 };
 
 int process_ssh_request(char *request);
+int process_pre_chroot_request(char **av);
 int winscp_regular_request(char *request);
 int winscp_transit_request(char *request);
 int process_winscp_requests(void);
@@ -160,15 +232,20 @@ int main (int argc, char **argv)
 #endif
 
 #ifdef UNIX_COMPAT 
-        openlog(PACKAGE_NAME, logopts, LOG_AUTH);
+	openlog(PACKAGE_NAME, logopts, LOG_AUTH);
 #elif IRIX_COMPAT
-        openlog(PACKAGE_NAME, logopts, LOG_AUTH);
+	openlog(PACKAGE_NAME, logopts, LOG_AUTH);
 #else
-        if (debuglevel > 1) /* debuglevel 1 will still log to syslog */
-                logopts |= LOG_PERROR;
-        openlog(PACKAGE_NAME, logopts, LOG_AUTHPRIV);
+	if (debuglevel > 1) /* debuglevel 1 will still log to syslog */
+		logopts |= LOG_PERROR;
+	openlog(PACKAGE_NAME, logopts, LOG_AUTHPRIV);
 #endif
 
+	if (debuglevel > 0)
+		debug = syslog;
+	else
+		debug = noop_syslog;
+
 #ifdef CHROOTED_NAME
 	/*
 	 *	is this a chroot'ed scponly installation?
@@ -181,8 +258,7 @@ int main (int argc, char **argv)
 	if (0==strncmp(argv[0],CHROOTED_NAME,FILENAME_MAX))
 #endif
 	{
-		if (debuglevel)
-			syslog(LOG_INFO, "chrooted binary in place, will chroot()");
+		debug(LOG_INFO, "chrooted binary in place, will chroot()");
 		chrooted=1;
 	}
 #endif /* CHROOTED_NAME */
@@ -194,11 +270,11 @@ int main (int argc, char **argv)
 		for (i=0;i<argc;i++)
 			syslog(LOG_DEBUG, "\targ %u is %s", i, argv[i]);
 	}
-        if (debuglevel)
+
 #ifdef UNIX_COMPAT
-                syslog(LOG_DEBUG, "opened log at LOG_AUTH, opts 0x%08x", logopts);
+	debug(LOG_DEBUG, "opened log at LOG_AUTH, opts 0x%08x", logopts);
 #else
-                syslog(LOG_DEBUG, "opened log at LOG_AUTHPRIV, opts 0x%08x", logopts);
+	debug(LOG_DEBUG, "opened log at LOG_AUTHPRIV, opts 0x%08x", logopts);
 #endif
 
 	if (getuid()==0)
@@ -213,21 +289,23 @@ int main (int argc, char **argv)
 	if (argc!=3)
 #endif
 	{
-		if (debuglevel)
-			syslog (LOG_ERR, "incorrect number of args");
+		debug(LOG_ERR, "incorrect number of args");
 		exit(EXIT_FAILURE);
 	}
 	if (!get_uservar())
 	{
-		syslog (LOG_ERR, "%s is misconfigured. contact sysadmin.", argv[0]);
+		syslog(LOG_ERR, "%s is misconfigured. contact sysadmin.", argv[0]);
 		exit (EXIT_FAILURE);
 	}
 
 #ifdef CHROOTED_NAME
 	if (chrooted)
 	{
+		char **av = NULL;
+		char *tmprequest = NULL;
 		char *root_dir = chrootdir;
 		char chdir_path[FILENAME_MAX];
+		
 
 		strcpy(chrootdir, homedir);
 		strcpy(chdir_path, "/");
@@ -236,6 +314,9 @@ int main (int argc, char **argv)
 			if (strncmp(root_dir, "//", 2) == 0) 
 			{
 				snprintf(chdir_path, FILENAME_MAX, "%s", root_dir + 1);
+				/* make sure HOME will be set to something correct if used*/
+				debug(LOG_DEBUG, "Setting homedir to %s", chdir_path);
+				strcpy(homedir, chdir_path);
 				*root_dir = '\0';
 				break;
 			}
@@ -269,35 +350,67 @@ int main (int argc, char **argv)
 			exit(EXIT_FAILURE);
 		}
 #endif
-		if (debuglevel)
-			syslog (LOG_DEBUG, "chrooting to dir: \"%s\"", chrootdir);
+
+/* already within CHROOTED_NAME block */
+#if defined(PASSWD_COMPAT) || defined(QUOTA_COMPAT)
+		
+		/*
+		 * perhaps we need to refactor so we don't have to exit right
+		 * in the middle of the code, but we can't chroot and expect to be
+		 * able to change the password and have it be of any use unless
+		 * there is some additional process that scponly is unaware of
+		 * happening on the back end.
+		 */
+		tmprequest = strdup(argv[2]);
+		av = build_arg_vector(tmprequest);
+		free(tmprequest);
+		if (
+#ifdef PASSWD_COMPAT
+			(exact_match(av[0],"passwd"))
+			|| (exact_match(av[0],PROG_PASSWD))
+#else
+			0
+#endif
+#ifdef QUOTA_COMPAT
+			|| (exact_match(av[0],"quota"))
+			|| (exact_match(av[0],PROG_QUOTA))
+#endif
+		) {
+			int status = process_pre_chroot_request(av);
+			discard_vector(av);
+			
+			if (status) {
+				syslog(LOG_ERR, "process_pre_chroot_request(%s) failed with code %i [%s]",
+					argv[2],WEXITSTATUS(status),logstamp());
+				exit(EXIT_FAILURE);
+			}
+			debug(LOG_DEBUG, "scponly completed");
+			exit(EXIT_SUCCESS);
+		} else {
+			discard_vector(av);
+		}
+
+#endif /* passwd or quota */
+
+		debug(LOG_DEBUG, "chrooting to dir: \"%s\"", chrootdir);
 		if (-1==(chroot(chrootdir)))
 		{
-			if (debuglevel)
-			{
-				syslog (LOG_ERR, "chroot: %m");
-			}
-			syslog (LOG_ERR, "couldn't chroot to %s [%s]", chrootdir, logstamp());
+			debug(LOG_ERR, "chroot: %m");
+			syslog(LOG_ERR, "couldn't chroot to %s [%s]", chrootdir, logstamp());
 			exit(EXIT_FAILURE);
 		}
-	    if (debuglevel)                                                                                                
-		{                                                                                                              
-			syslog (LOG_DEBUG, "chdiring to dir: \"%s\"", chdir_path);                                             
-		}                                                                                                              
-		if (-1==(chdir(chdir_path)))                                                                                   
-		{                                                                                                              
-			if (debuglevel)                                                                                        
-			{                                                                                                      
-				syslog (LOG_ERR, "chdir: %m");                                                                 
-			}                                                                                                      
-			syslog (LOG_ERR, "couldn't chdir to %s [%s]", chdir, logstamp());                                      
+		
+		debug(LOG_DEBUG, "chdiring to dir: \"%s\"", chdir_path);					     
+		if (-1==(chdir(chdir_path)))										   
+		{													      
+			debug(LOG_ERR, "chdir: %m");								 
+			syslog (LOG_ERR, "couldn't chdir to %s [%s]", chdir_path, logstamp());				      
 			exit(EXIT_FAILURE);     
 		}
 	}
 #endif /* CHROOTED_NAME */
 
-	if (debuglevel)
-		syslog (LOG_DEBUG, "setting uid to %u", getuid());
+	debug(LOG_DEBUG, "setting uid to %u", getuid());
 	if (-1==(seteuid(getuid())))
 	{
 		syslog (LOG_ERR, "couldn't revert to my real uid. seteuid: %m");
@@ -307,8 +420,7 @@ int main (int argc, char **argv)
 #ifdef WINSCP_COMPAT
 	if (argc==1)
 	{
-		if (debuglevel)
-			syslog (LOG_DEBUG, "entering WinSCP compatibility mode [%s]",logstamp());
+		debug(LOG_DEBUG, "entering WinSCP compatibility mode [%s]",logstamp());
 		if (-1==process_winscp_requests())
 		{
 			syslog(LOG_ERR, "failed WinSCP compatibility mode [%s]", logstamp());
@@ -323,11 +435,90 @@ int main (int argc, char **argv)
 		syslog(LOG_ERR, "bad request: %s [%s]", argv[2], logstamp());
 		exit(EXIT_FAILURE);
 	}
-	if (debuglevel)
-		syslog(LOG_DEBUG, "scponly completed");
+	debug(LOG_DEBUG, "scponly completed");
 	exit(EXIT_SUCCESS);
 }
 
+#ifdef CHROOTED_NAME
+#if defined(PASSWD_COMPAT) || defined(QUOTA_COMPAT)
+
+int process_pre_chroot_request(char ** av) {
+
+	char *flat_request = NULL;
+	char *env[2] = { NULL, NULL };
+	int retval = -1;
+	
+	/* revert to real user so I'm not changing the passwd as root */
+	debug(LOG_DEBUG, "handling pre-chroot request");
+	debug(LOG_DEBUG, "setting uid to %u", getuid());
+
+	if (-1==(seteuid(getuid())))
+	{
+		syslog (LOG_ERR, "couldn't revert to my real uid. seteuid: %m");
+		exit(EXIT_FAILURE);
+	}
+
+	av[0] = substitute_known_path(av[0]);
+	flat_request = flatten_vector(av);
+	
+	/* 
+	 * sanity check, substitute_known_path should have given the exact path,
+	 * ONLY execute if an exact match was found
+	 */
+	if (
+#ifdef PASSWD_COMPAT
+		(!exact_match(av[0], PROG_PASSWD))
+#else
+		1
+#endif
+#ifdef QUOTA_COMPAT
+		&& (!exact_match(av[0], PROG_QUOTA))
+#endif
+	) {
+		syslog(LOG_ERR, "Invalid pre-chroot request attempted: '%s' [%s]", av[0], logstamp());
+		exit(EXIT_FAILURE);
+	}
+	
+	if(check_dangerous_args(av))
+	{
+		syslog(LOG_ERR, "requested command (%s) tried to use disallowed argument [%s])", 
+			flat_request, logstamp());
+		exit(EXIT_FAILURE);
+	}
+
+	if (valid_arg_vector(av))
+	{
+		int status = 0;
+		
+		debug(LOG_DEBUG, "about to fork (%s) [%s]",flat_request,logstamp());
+		
+		if (fork() == 0)
+			retval=execve(av[0],av,env);
+		else
+		{
+			wait(&status);
+			fflush(stdout);
+			fflush(stderr);
+			free(flat_request);
+			
+			debug(LOG_DEBUG, "forked child returned... [%s]",logstamp());
+			
+			return WEXITSTATUS(status);
+		}
+
+	} else {
+		syslog(LOG_ERR, "invalid argument vector (%s) [%s]", 
+			flat_request,logstamp());
+		free(flat_request);
+		
+		exit(EXIT_FAILURE);
+	}
+	return retval;
+}
+
+#endif /* passwd or quota support */
+#endif /* chrooted support */
+
 #ifdef WINSCP_COMPAT
 
 int winscp_transit_request(char *request)
@@ -403,7 +594,7 @@ int winscp_regular_request(char *request)
 
 int process_winscp_requests(void)
 {
-        char    linebuf[MAX_REQUEST];
+	char    linebuf[MAX_REQUEST];
 	int	count=0;		/* num of semicolons in cmd */
 	int	ack=0;
 
@@ -450,15 +641,14 @@ int process_winscp_requests(void)
 
 int process_ssh_request(char *request)
 {
-	char **av, **tenv;
+	char **av, **tmp_av, **tenv;
 	char *flat_request,*tmpstring, *tmprequest;
 	char bad_winscp3str[] = "test -x /usr/lib/sftp-server && exec /usr/lib/sftp-server test -x /usr/local/lib/sftp-server && exec /usr/local/lib/sftp-server exec sftp-server";
 	int retval;
 	int reqlen=strlen(request);
-	char *env[2] = { NULL, NULL };
+	char **env = NULL;
 
-	if (debuglevel)
-		syslog(LOG_DEBUG, "processing request: \"%s\"\n", request);
+	debug(LOG_DEBUG, "processing request: \"%s\"\n", request);
 
 	tmprequest=strdup(request);
 
@@ -466,7 +656,7 @@ int process_ssh_request(char *request)
 
 	bad_winscp3str[57]=10;
 	bad_winscp3str[127]=10;
-
+			
 	if(strcmp(request,bad_winscp3str)==0)
 	{
 	    /*
@@ -478,7 +668,7 @@ int process_ssh_request(char *request)
 	}
 #endif
 
-
+	
 #ifdef GFTP_COMPAT 
 	/*
 	 *	gFTP compatibility hack
@@ -498,38 +688,38 @@ int process_ssh_request(char *request)
 	 */
 	if (!valid_chars(tmprequest))
 	{
+		debug(LOG_DEBUG, "rejected because of invalid chars (%s)", logstamp());
 		free(tmprequest);
 		return(-1);
 	}
 #endif
 
 #ifdef WINSCP_COMPAT
-        if (strbeg(tmprequest,PROG_CD))
-        {
-                char *destdir=(char *)malloc(reqlen);
-                if (destdir == NULL)
-                {
-                        perror("malloc");
-                        exit(EXIT_FAILURE);
-                }
+	if (strbeg(tmprequest,PROG_CD))
+	{
+		char *destdir=(char *)malloc(reqlen);
+		if (destdir == NULL)
+		{
+			perror("malloc");
+			exit(EXIT_FAILURE);
+		}
 
-                /*
-                 * well, now that scponly is a persistent shell
-                 * i have to maintain a $PWD.  damn.
-                 * we're going to INSIST upon a double quote
-                 * encapsulated new directory to change to.
-                 */
-                if ((tmprequest[(reqlen-1)]=='"') && (tmprequest[3]=='"'))
-                {
-                        bzero(destdir,reqlen);
-                        strncpy(destdir,&tmprequest[4],reqlen-5);
-                        if (debuglevel)
-                                syslog(LOG_INFO, "chdir: %s (%s)", tmprequest, logstamp());
-                        retval=chdir(destdir);
-                        free(destdir);
-						free(tmprequest);
-                        return(retval);
-                }
+		/*
+		 * well, now that scponly is a persistent shell
+		 * i have to maintain a $PWD.  damn.
+		 * we're going to INSIST upon a double quote
+		 * encapsulated new directory to change to.
+		 */
+		if ((tmprequest[(reqlen-1)]=='"') && (tmprequest[3]=='"'))
+		{
+			bzero(destdir,reqlen);
+			strncpy(destdir,&tmprequest[4],reqlen-5);
+			debug(LOG_INFO, "chdir: %s (%s)", tmprequest, logstamp());
+			retval=chdir(destdir);
+			free(destdir);
+			free(tmprequest);
+			return(retval);
+		}
 		syslog(LOG_ERR, "bogus chdir request: %s (%s)", tmprequest, logstamp());
 		free(tmprequest);
 		return(-1);
@@ -545,7 +735,7 @@ int process_ssh_request(char *request)
 	 * clean any path info from request and substitute our known pathnames
 	 */
 	av[0] = substitute_known_path(av[0]);
-
+	
 	/*
 	 * we only process wildcards for scp commands
 	 */
@@ -566,23 +756,31 @@ int process_ssh_request(char *request)
 		chdir(DEFAULT_CHDIR);
 	}
 #endif
+	
 
 	flat_request = flatten_vector(av);
 
-	if(check_dangerous_args(av))
+	/* 
+	 * Use a temp arg vector since getopt will permute the command line arguments
+	 * for anything that it does not know about.  If all rsync options are well
+	 * defined this isn't necessary.
+	 */
+	tmp_av = build_arg_vector(flat_request);
+	if(check_dangerous_args(tmp_av))
 	{
 		syslog(LOG_ERR, "requested command (%s) tried to use disallowed argument (%s))", 
 			flat_request, logstamp());
 		exit(EXIT_FAILURE);
 	}
+	discard_vector(tmp_av);
 
 	if (valid_arg_vector(av))
 	{
 
-/*                                                                                                                   
- * Unison needs the HOME environment variable be set to the directory                                                  
- * where the .unison directory resides.                                                                                
- */                                                                                                                    
+/*														   
+ * Unison needs the HOME environment variable be set to the directory						  
+ * where the .unison directory resides.										
+ */														    
 #ifdef USE_SAFE_ENVIRONMENT
 		safeenv[0] = NULL;
 		filter_allowed_env_vars();
@@ -592,29 +790,20 @@ int process_ssh_request(char *request)
 				syslog(LOG_DEBUG, "Environment contains \"%s\"", *tenv++);
 			}
 		}
+		env = safeenv;
 #endif
 
 #ifdef UNISON_COMPAT
-		if (chrooted)                                                                                           
-		{                                                                                                       
-			if (!mysetenv("HOME","/"))                                                                       
-			{                                                                                               
-				syslog(LOG_ERR, "could not set HOME environment variable to '/' (%s)", logstamp());           
-				exit(EXIT_FAILURE);                                                                     
-			}                                                                                               
-			if (debuglevel)                                                                                 
-				syslog(LOG_DEBUG, "set HOME environment variable to / %s", logstamp());           
-		}                                                                                                       
-		else
+		/* the HOME environment variable should have been set above, but I need to make sure
+		 * that it's value as read from the environment is replaced with the actual value
+		 * as it exists within the chroot, which is what the applications will expect to see.
+		 */
+		if (replace_env_entry("HOME",homedir) && (((strlen(homedir) + 6 ) > FILENAME_MAX) || !mysetenv("HOME",homedir)))
 		{
-			if (((strlen(homedir) + 6 ) > FILENAME_MAX) || !mysetenv("HOME",homedir))
-			{
-				syslog(LOG_ERR, "could not set HOME environment variable (%s)", logstamp());
-				exit(EXIT_FAILURE);
-			}
-			if (debuglevel)
-				syslog(LOG_DEBUG, "set HOME environment variable to %s (%s)", homedir, logstamp());
+			syslog(LOG_ERR, "could not set HOME environment variable (%s)", logstamp());
+			exit(EXIT_FAILURE);
 		}
+		debug(LOG_DEBUG, "set non-chrooted HOME environment variable to %s (%s)", homedir, logstamp());
 #endif 
 		syslog(LOG_INFO, "running: %s (%s)", flat_request, logstamp());
 
@@ -623,19 +812,15 @@ int process_ssh_request(char *request)
 		{
 			int status=0;
 			if (fork() == 0)
-#ifdef USE_SAFE_ENVIRONMENT
-				retval=execve(av[0],av,safeenv);
-#else
-				retval=execve(av[0],av,NULL);
-#endif
+				retval=execve(av[0],av,env);
 			else
 			{
 				wait(&status);
 				fflush(stdout);
 				fflush(stderr);
 				discard_vector(av);
-#ifdef USE_SAFE_ENVIRONMENT                                                                                            
-				discard_vector(safeenv);
+#ifdef USE_SAFE_ENVIRONMENT											    
+				discard_child_vectors(safeenv);
 #endif
 				free(flat_request);
 				free(tmprequest);
@@ -645,17 +830,14 @@ int process_ssh_request(char *request)
 		else
 #endif
 		{
-#ifdef USE_SAFE_ENVIRONMENT
-			retval=execve(av[0],av,safeenv);
-#else
-			retval=execve(av[0],av,NULL);
-#endif
+			debug(LOG_DEBUG, "about to exec \"%s\" (%s)", av[0], logstamp());
+			retval=execve(av[0],av,env);
 		}
 		syslog(LOG_ERR, "failed: %s with error %s(%u) (%s)", flat_request, strerror(errno), errno, logstamp());
 		free(flat_request);
 		discard_vector(av);
-#ifdef USE_SAFE_ENVIRONMENT                                                                                            
-		discard_vector(safeenv);
+#ifdef USE_SAFE_ENVIRONMENT
+		discard_child_vectors(safeenv);
 #endif
 #ifdef WINSCP_COMPAT
 		if (winscp_mode)
@@ -692,3 +874,4 @@ int process_ssh_request(char *request)
 		exit(EXIT_FAILURE);
 }
 
+/* vim: set noet sw=4 ts=4: */
